/*
 * @file purple-io.c  Buffering I/O layer.
 * @ingroup core
 *
 * Pidgin is the legal property of its developers, whose names are too numerous
 * to list here.  Please refer to the COPYRIGHT file distributed with this
 * source distribution.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "internal.h"
#include "eventloop.h"
#include "purple-io.h"

#define OPERATION_MAX_BYTES 4096
#define WRITE_LOW_WATER     512

static void update_watches (PurpleIO *io);

static gboolean
can_read (PurpleIO *io)
{
	fd_set rfds;
	struct timeval tv;
	int retval;

	FD_ZERO(&rfds);
	FD_SET(io->fd, &rfds);

	tv.tv_sec = 0;
	tv.tv_usec = 1;

	retval = select (io->fd + 1, &rfds, NULL, NULL, &tv);

	return retval ? TRUE : FALSE;
}

static gint
try_read (PurpleIO *io, gint max_read, gint *read_errno)
{
	guint8 buf [OPERATION_MAX_BYTES];
	glong  flags;
	gint   n_read;

	/* Do a non-blocking read */
	flags = fcntl (io->fd, F_GETFL);
	fcntl (io->fd, F_SETFL, flags | O_NONBLOCK);

	if (io->gsc) {
		n_read = purple_ssl_read_nonblock (io->gsc, buf, MIN (max_read, OPERATION_MAX_BYTES));

		if (read_errno) {
			if (n_read == PURPLE_SSL_IO_AGAIN)
				*read_errno = EAGAIN;
			else if (n_read == PURPLE_SSL_IO_INTERRUPTED)
				*read_errno = EINTR;
			else
				*read_errno = 0;
		}

		if (n_read < 0)
			n_read = 0;
 	} else {
		n_read = read (io->fd, buf, MIN (max_read, OPERATION_MAX_BYTES));
		if (read_errno)
			*read_errno = errno;
	}

	fcntl (io->fd, F_SETFL, flags);

	if (n_read > 0) {
		/* Got some data */
		purple_fifo_enqueue (io->read_fifo, buf, n_read);
	}

	return n_read;
}

static void
read_nonblock (PurpleIO *io, gint max_read)
{
	gint read_errno;

	if (io->got_eof)
		return;

	if (max_read <= purple_fifo_get_used (io->read_fifo))
		return;

	max_read -= purple_fifo_get_used (io->read_fifo);

	if (can_read (io) && try_read (io, max_read, &read_errno) < 1 && read_errno != EINTR)
	{
		/* Disconnected */
		io->got_eof = TRUE;
		update_watches (io);

		if (io->eof_func)
			io->eof_func (io, io->eof_func_data);
	}
}

static void
read_cb (gpointer data, gint source, PurpleInputCondition condition)
{
	PurpleIO *io = data;
	gint    n_read;
	gint    read_errno;

	n_read = try_read (io, OPERATION_MAX_BYTES, &read_errno);

	/* If interrupted by signal, just retry later */
	if (n_read < 0 && read_errno == EINTR)
		return;

	/* If we got 0 bytes or an error that's not EINTR, the connection was closed */
	if (n_read < 1) {
		io->got_eof = TRUE;
		update_watches (io);

		if (io->eof_func)
			io->eof_func (io, io->eof_func_data);

		return;
	}

	if (io->read_func)
		io->read_func (io, io->read_func_data);

	update_watches (io);
}

static void
write_cb (gpointer data, gint source, PurpleInputCondition condition)
{
	PurpleIO *io = data;
	guint8  buf [OPERATION_MAX_BYTES];
	gint    n_written;
	glong   flags;

	n_written = purple_fifo_get_used (io->write_fifo);
	n_written = MIN (OPERATION_MAX_BYTES, n_written);

	/* Paranoid: Shouldn't happen */
	if (n_written < 1)
		return;

	purple_fifo_peek (io->write_fifo, buf, n_written);

	flags = fcntl (io->fd, F_GETFL);
	fcntl (io->fd, F_SETFL, flags | O_NONBLOCK);

	if (io->gsc) {
		n_written = purple_ssl_write_nonblock (io->gsc, buf, n_written);
 	} else {
		n_written = write (io->fd, buf, n_written);
	}

	fcntl (io->fd, F_SETFL, flags);

	if (n_written < 0)
		io->got_eof = TRUE;

	if (n_written < 1)
		return;

	/* Drop written data from buffer */
	purple_fifo_dequeue (io->write_fifo, NULL, n_written);

	if (io->flushed_func &&
	    purple_fifo_get_used (io->write_fifo) == 0)
		io->flushed_func (io, io->flushed_func_data);

	if (io->write_func &&
	    purple_fifo_get_used (io->write_fifo) < WRITE_LOW_WATER)
		io->write_func (io, io->write_func_data);

	update_watches (io);
}

static void
update_watches (PurpleIO *io)
{
	gboolean want_read_watch  = FALSE;
	gboolean want_write_watch = FALSE;

	if (!io->got_eof) {
		if (io->read_func)
			want_read_watch = TRUE;

		if (purple_fifo_get_used (io->write_fifo))
			want_write_watch = TRUE;
	}

	if (want_read_watch && !io->read_watch_id) {
		io->read_watch_id = purple_input_add (io->fd, PURPLE_INPUT_READ, read_cb, io);
	} else if (!want_read_watch && io->read_watch_id) {
		purple_input_remove (io->read_watch_id);
		io->read_watch_id = 0;
	}

	if (want_write_watch && !io->write_watch_id) {
		io->write_watch_id = purple_input_add (io->fd, PURPLE_INPUT_WRITE, write_cb, io);
	} else if (!want_write_watch && io->write_watch_id) {
		purple_input_remove (io->write_watch_id);
		io->write_watch_id = 0;
	}
}

PurpleIO *
purple_io_new (gint fd)
{
	PurpleIO *io;

	io = g_new0 (PurpleIO, 1);

	io->fd         = fd;
	io->read_fifo  = purple_fifo_new ();
	io->write_fifo = purple_fifo_new ();

	return io;
}

PurpleIO *
purple_io_new_ssl (PurpleSslConnection *gsc)
{
	PurpleIO *io;

	io = g_new0 (PurpleIO, 1);

	io->gsc        = gsc;
	io->fd         = gsc->fd;
	io->read_fifo  = purple_fifo_new ();
	io->write_fifo = purple_fifo_new ();

	return io;
}

void
purple_io_destroy (PurpleIO *io)
{
#if 0
	if (purple_fifo_get_used (io->write_fifo) > 0)
		g_print ("Throwing away outbound data!\n");
#endif

	if (io->read_watch_id)
		purple_input_remove (io->read_watch_id);
	if (io->write_watch_id)
		purple_input_remove (io->write_watch_id);

	purple_fifo_destroy (io->read_fifo);
	purple_fifo_destroy (io->write_fifo);

	if (io->gsc)
		purple_ssl_close (io->gsc);
	else
		close (io->fd);

	g_free (io);
}

gint
purple_io_get_fd (PurpleIO *io)
{
	return io->fd;
}

PurpleSslConnection *
purple_io_get_ssl_connection (PurpleIO *io)
{
	return io->gsc;
}

gboolean
purple_io_is_connected (PurpleIO *io)
{
	return !io->got_eof;
}

void
purple_io_set_read_func (PurpleIO *io, PurpleIOFunc func, gpointer data)
{
	io->read_func      = func;
	io->read_func_data = data;

	update_watches (io);
}

void
purple_io_set_write_func (PurpleIO *io, PurpleIOFunc func, gpointer data)
{
	io->write_func      = func;
	io->write_func_data = data;

	update_watches (io);
}

void
purple_io_set_eof_func (PurpleIO *io, PurpleIOFunc func, gpointer data)
{
	io->eof_func      = func;
	io->eof_func_data = data;

	update_watches (io);
}

void
purple_io_set_flushed_func (PurpleIO *io, PurpleIOFunc func, gpointer data)
{
	io->flushed_func      = func;
	io->flushed_func_data = data;

	update_watches (io);
}

gboolean
purple_io_write_bytes (PurpleIO *io, gconstpointer libpurple, gint n_bytes)
{
	if (io->got_eof)
		return FALSE;

	purple_fifo_enqueue (io->write_fifo, libpurple, n_bytes);

	update_watches (io);
	return TRUE;
}

gboolean
purple_io_write_line (PurpleIO *io, const gchar *line)
{
	if (io->got_eof)
		return FALSE;

	purple_fifo_enqueue (io->write_fifo, line, strlen (line));
	purple_fifo_enqueue (io->write_fifo, "\r\n", 2);

	update_watches (io);
	return TRUE;
}

gboolean
purple_io_read_bytes (PurpleIO *io, gpointer dest, gint n_bytes)
{
	gboolean result;

	read_nonblock (io, n_bytes);

	result = purple_fifo_dequeue (io->read_fifo, dest, n_bytes);

	update_watches (io);
	return result;
}

gboolean
purple_io_read_line (PurpleIO *io, gchar **line)
{
	gboolean result;

	read_nonblock (io, 65536);

	result = purple_fifo_dequeue_line (io->read_fifo, line);

	update_watches (io);
	return result;
}

gint
purple_io_get_pending_read (PurpleIO *io)
{
	return purple_fifo_get_used (io->read_fifo);
}

gint
purple_io_get_pending_write (PurpleIO *io)
{
	return purple_fifo_get_used (io->write_fifo);
}
